#Consensus.pm
package Consensus;

#@EXPORT = qw(
#	project_constraint_and_fold_seq
#	);
################################################################################
## A perl library with plenty of useful commands to manipulate RNA alignmnets and
## consensus structures
## It uses RNAtools library
################################################################################

#use feature ':5.10';
use strict 'vars';
use warnings;
use Getopt::Long;
use Pod::Usage;
use List::Util qw/ min max /;
#use POSIX qw/ceil floor/;
use Cwd qw/abs_path getcwd/;



#use FindBin;                 # locate this script
#use lib "$FindBin::Bin/../lib/";  # use the parent directory


use StructureLibrary::Tools;
use StructureLibrary::Sequence;

use StructureLibrary::Structure;

use MultipleAlignment::Tools;

sub LOAD_CONFIG {
	my $local_config_path = shift;
	StructureLibrary::Structure::LOAD_CONFIG($local_config_path);
}

###############################################################################
# METHODS
###############################################################################
#######################################################################
# Convert sequence positions to N 
# INPUT:
# OUTPUT:
#######################################################################
sub convert_covar_bppos_to_N {
	my ($seq_with_gap, $seq, $covstruct, $foldedstruct)= @_;
	if (length($seq_with_gap) != length($covstruct)) {
		die "Expected same length of gapped sequence and covarying-struct\n $seq_with_gap\n $covstruct\n";
	}

	if (length($seq) != length($foldedstruct)) {
		die "Expected same length of sequence and folded-struct\n $seq\n $foldedstruct\n";
	}

	my @seq_arr = split '',$seq;
	my @gapped_seq_arr = split '',$seq_with_gap;
	my @gapped_Ned_seq_arr = split '',$seq_with_gap;
	my @covstruct_arr = split '',$covstruct;
	my @foldedstruct_arr = split '',$foldedstruct;
	for my $ii (0 .. (length($covstruct)-1)) {
		#print "covstruct_arr[$ii]:". $covstruct_arr[$ii]."\n";
		if(($covstruct_arr[$ii] eq '(') or ($covstruct_arr[$ii] eq ')') or ($covstruct_arr[$ii] eq '<') or ($covstruct_arr[$ii] eq '>')) {
			if ($gapped_Ned_seq_arr[$ii] =~ /[a-zA-Z]/) {
				$gapped_Ned_seq_arr[$ii] = 'N';
			}	
		} 
	}

	#print "cov structu       $covstruct\n" ;
	my $Ned_seq = join('', @gapped_Ned_seq_arr);
	#print "gapped_Ned_seq1   $Ned_seq\n";
	$Ned_seq =~ s/[^a-zA-Z]+//g;
	print "ungapped_Ned_seq1 $Ned_seq\n";

	my @Ned_seq_arr = split '',$Ned_seq;
	#print "before reverse    " . join('', @gapped_Ned_seq_arr) . "\n";
	for my $ii (0 .. (length($foldedstruct)-1)) { # Reverse back Ned positions if they do not occur in mfe folded structure
		if( ($Ned_seq_arr[$ii] eq 'N') and  
			!(($foldedstruct_arr[$ii] eq '(') or ($foldedstruct_arr[$ii] eq ')') or ($foldedstruct_arr[$ii] eq '<') or ($foldedstruct_arr[$ii] eq '>'))
			)
		{
			$Ned_seq_arr[$ii] = $seq_arr[$ii];
		}
	}
	my $Ned_seq_pruned = join('', @Ned_seq_arr);

	print "Ned_seq_pruned    $Ned_seq_pruned\n";
	print "foldedstruct      $foldedstruct\n";
	my $Ned_pos_count = $Ned_seq_pruned =~ tr/N//;
	
	print "#INFO: Ned_seq_pruned_len: " . length($Ned_seq_pruned) . " Ned_pos_count: $Ned_pos_count \n";
	print "Ned_seq_pruned2   $Ned_seq_pruned\n";

	#print "seq:              $seq\n";
	die "Error: Expected same length of ungapped sequence and ned-seq\n $seq\n $Ned_seq\n" if (length($seq) != length($Ned_seq));

	return $Ned_seq_pruned;
	
}


##############
#map_cons_fold_seq($consensus, @{$seqs_aref}[$elem_count-1], $elem, $i_cutoff_cons, $i_cutoff_mfe);
##method
sub  project_constraint_and_fold_seq
{
  
  my ($constraint, $sequence, $elem, $elem_count, $rfam_id, $mapping_cutoff_ratio, $bps_cutoff_mfe, $stdout, 
  $RESULTDIR, $TMPDIR, $covarying_bp_struct ) = @_;
  print "Calling project_constraint_and_fold_seq id:$rfam_id\n";	 
  ## map consensus structure to single element
  my ($structure, $seq, $ratio) = MultipleAlignment::Tools::map_consensus_structure_to_single_alignment_sequence(
							  $constraint, $sequence);

  if($ratio == 0)
  {
	  print STDERR "WARNING: $elem: no structure left, the consensus structure is entirely incompatible!\n";
  }
  elsif ($ratio == -1)
  {
	  print STDERR "WARNING: $elem: no basepairs in the consensus structure!\n";
  }
  if ($ratio*100 < $mapping_cutoff_ratio && $ratio != -1)
  {
	  print STDERR "$elem: skipped, constraint; less than threshold mapped_constraint/consensus_struct remains! (".sprintf("%.2f",$ratio).")\n";
	  return "";
  }

  my $elem_line = $elem."\t";


  ## start constraint folding with RNAfold
  if ($seq && $structure)
  {
	  
	  ## create constraint fasta file
	  mkdir($RESULTDIR) unless(-d $RESULTDIR );
	  my $nice_elem = $elem;
	  $nice_elem =~ tr/\//_/;
	  my @constraint_fasta = (">$nice_elem", $seq, $structure." #S");
	  my $constraint_fasta_file = $TMPDIR."/constraint.fasta";
	  StructureLibrary::Tools::write_file_from_array($constraint_fasta_file, \@constraint_fasta);
	  print "nice_elem: $nice_elem\n"; 
	  ## create sequence only fasta file
	  my @sequence_fasta = (">$nice_elem", $seq);
	  my $tmp_fasta_name = $nice_elem;
	  $tmp_fasta_name =~ s/[^A-Za-z0-9\-\.]//g;  # Clenaup invalid characters for filename
	  $tmp_fasta_name =~ s/[\.]+/\./g;  # Remove consecutive dots

	  my $seq_fasta_file  = $TMPDIR."/".$tmp_fasta_name.".fa";
	  print "seq_fasta_file is $seq_fasta_file\n";
	  StructureLibrary::Tools::write_file_from_array($seq_fasta_file, \@sequence_fasta);

	  # the main task of constraint folding is done by this function:
	  # the sequence is folded twice, with&without constraints, then the nergy diffrences and so on will be obsereved 
	  my $rna_href = ensemble_energies_and_probability_with_constraint(
					  $seq_fasta_file, $constraint_fasta_file, "-p -d 2 --noLP", 
					  $TMPDIR, 1);
					  #toask: what does the probabilty of structure mean? (Ens_E-STR_E)/RT
		  
  


	  ## .write all info to .struct file

	  
  #			die ("testing: elem line is now $elem_line\n");
	  
	  ##counting and comparing basepair counts inf mfe and constraint folded strcutures
	  my $cons_bp_href = StructureLibrary::Structure::structure2basepairs($structure);
	  my $mfe_bp_href = StructureLibrary::Structure::structure2basepairs($rna_href->{CONSTRAINT_MFE});
	  my @cons_bp = keys (%{$cons_bp_href});
	  my $total = @cons_bp;
	  my $bp_in_mfe = 0;
	  foreach (@cons_bp){
		  if(exists($mfe_bp_href->{$_})){
			  ++$bp_in_mfe;
		  } 
	  }

	  if ($total > 0 and $bp_in_mfe/$total < $bps_cutoff_mfe/100)
	  {
		  print STDERR "$elem: skipped, MFE strcuture; less than threshold ratio on MFE/constraint_structure remains!(". sprintf("%.2f",$bp_in_mfe/$total) .")\n";
		  return "";
	  }
	  
	  ## Write all seqs+structures to file rfam_id.fa or stdout
	  if ($stdout)
	  {
		  print STDOUT join("\n",@constraint_fasta)."\n";	
	  }
	  else
	  {
		  ## Append sequence+structure to a common fasta file
		  my $append = 1;
		  my $modified_seq = $seq;
		  ##folded seqeunce, mfe strcuture under constraints
		  if ($covarying_bp_struct) {
		  	#print "sequence $sequence\nseq $seq\ncovarying_bp_struct $covarying_bp_struct\n";
		  	$modified_seq = convert_covar_bppos_to_N($sequence, $seq, $covarying_bp_struct,  $rna_href->{CONSTRAINT_MFE});
		  	
		  }
		  else{
		  	$modified_seq = $seq;
		  }
		  my @folded_fasta = (">$nice_elem-covbps", $modified_seq, $rna_href->{CONSTRAINT_MFE}." #S");
		  if ($elem_count > 1) #append the following elements
		  {
			  ## append folded structure	
			  StructureLibrary::Tools::write_file_from_array(abs_path($RESULTDIR."/$rfam_id.folded.fa"), \@folded_fasta, $append); 
		  }
		  else # create a new file for the first element
		  {
			  ## write folded structure	
			  StructureLibrary::Tools::write_file_from_array(abs_path($RESULTDIR."/$rfam_id.folded.fa"), \@folded_fasta); 
			    
		  }												
	  }

	  ## add ratio  
	  $elem_line .= "$ratio\t";
	  
	  if($total == 0){
		  $elem_line .= "-\t";
	  }
	  else
	  {
		  $elem_line .= ($bp_in_mfe/$total)."\t";
	  }
	  ## add consensus structure to line
	  $elem_line .= "\t$structure\t";
	  ## add folded structure to line			
	  $elem_line .= $rna_href->{CONSTRAINT_MFE}."\t";

	  $elem_line .= $rna_href->{ENSEMBLE_E}."\t";
	  $elem_line .= $rna_href->{CONSTRAINT_E}."\t";
	  $elem_line .= $rna_href->{CONSTRAINT_MFE_E}."\t";
	  $elem_line .= $rna_href->{CONSTRAINT_P}."\t";
	  $elem_line .= $rna_href->{CONSTRAINT_MFE_P}."\t";
	  if($rna_href->{CONSTRAINT_P}){
		  $elem_line .= (($rna_href->{CONSTRAINT_P} - $rna_href->{CONSTRAINT_MFE_P}) / $rna_href->{CONSTRAINT_P});
	  } else {
		  $elem_line .= "-";
	  }
  } else {
	  $elem_line .= "0\t0\t0\t0\t0\t0\t0\t0\t0\t0";
  }

  return $elem_line;
}


sub parse_clustal_href{
	my $file = shift;
	my $FUNCTION = "parse_clustal in consensus_to_single_structures.pl";
	
	open(IN_HANDLE, "<$file") || die "ERROR in $FUNCTION:\n Couldn't open file $file/n";
	
	my %alignments = ();
	while(my $line = <IN_HANDLE>){
		chomp($line);
		
		# get entry line and add it to alignments hash
#		if ($line =~ /^(\S+\/\d+-\d+)\s+(\S+)$/){ Rfam based headers is not necessary 
		if ($line =~ /^(\S+)\s+(\S+)$/){
			if(exists ($alignments{$1})){
				$alignments{$1} .= $2;
			} else {
				$alignments{$1} = $2;
			}
		}
	}
	return \%alignments;
}
1;
